# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from enum import IntEnum

import os
from io import BytesIO
from json import loads

import requests
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.exceptions import MultipleObjectsReturned, PermissionDenied, ValidationError
from django.core.mail import EmailMessage
from django.core.paginator import Paginator, PageNotAnInteger, EmptyPage
from django.core.validators import validate_email
from django.db.models import Q
from django.http import Http404, HttpResponse, HttpResponseForbidden, JsonResponse
from django.shortcuts import render, redirect, get_object_or_404
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils import timezone
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods
from django_sendfile import sendfile
from paypal.standard.forms import PayPalPaymentsForm, PayPalEncryptedPaymentsForm

from accounts.models import User
from accounts.utils import get_ldap_user
from guichet.decorators import manager_required, manager_protected
from guichet.filters import RegistrationUnpaidFilter, RegistrationPaidFilter
from guichet.forms import (
    PaypalAccountForm,
    EventForm,
    OpenTicketForm,
    RestrictedTicketForm,
    GodsonForm,
    RegistrationForm,
    UserManagerForm,
    GroupManagerForm,
    CheckForm,
    TermOfUseForm,
    ChoiceOptionsForm,
    ChoiceOptionsWithGroupForm,
    InfoSupCreationForm,
    InfoSupChoicesSetCreationFormSet,
    InfoSupChoicesSetEditFormSet,
    InfoSupReceivedForm,
    OptionAloneCreationForm,
    OptionGroupCreationForm,
    OptionGroupSetCreationFormSet,
    OptionGroupSetEditFormSet,
    LydiaAccountForm,
    LydiaPaymentForm,
    LydiaTokenForm,
)
from guichet.models import (
    Event,
    Ticket,
    Manager,
    UserManager,
    GroupManager,
    Registration,
    Option,
    LogEntry,
    Info_Sup,
    Info_Sup_Recv,
    Info_Sup_Recv_File,
)
from guichet.utils import (
    mail_confirm_registration,
    get_event_stats,
    check_registration,
    check_registration_sandbox,
    list_registrations,
    maj_state_registration,
)
from guichet.utils.event_export import Event_export
from guichet.utils.registration import registration_to_image, print_sandbox_billet


def home(request):
    return redirect(reverse("list-events"))


@login_required
def paypal_account_create(request):
    if not request.user.has_perm("guichet.add_event"):
        raise PermissionDenied

    form = PaypalAccountForm(request.POST or None)

    if request.method == "POST" and form.is_valid():
        form.save()
        messages.success(request, "Compte Paypal enregistré avec succès.")

        return redirect(reverse("add-event"))

    c = {
        "form": form,
    }

    return render(request, "guichet/paypal_account_create.html", c)


@login_required
def lydia_account_create(request):
    if not request.user.has_perm("guichet.add_event"):
        raise PermissionDenied

    form = LydiaAccountForm(request.POST or None)

    if request.method == "POST" and form.is_valid():
        # API vendortoken: vérifier la validité des tokens
        url = settings.LYDIA_API_URL + "/auth/vendortoken.json"
        data = {
            "vendor_token": form.cleaned_data.get("vendor_token"),
            "vendor_id": form.cleaned_data.get("private_token"),
        }

        response = requests.post(url, data)
        if response.status_code == 200:
            content = loads(response.content)

            if content.get("status") == "ok":  # Tokens valides
                form.save()
                messages.success(request, "Compte Lydia enregistré avec succès.")
                return redirect(reverse("add-event"))

            else:
                messages.error(request, f"Erreur : {content.get('message')}")

        else:
            messages.error(request, f"Erreur lors l'envoie de la requête : code {response.status_code}")

    c = {
        "form": form,
    }

    return render(request, "guichet/lydia_account_create.html", c)


@login_required
def lydia_get_token(request):
    if not request.user.has_perm("guichet.add_event"):
        raise PermissionDenied

    if request.method == "POST" and "account_create" in request.POST:
        form = LydiaAccountForm(request.POST)
        form.save()
        messages.success(request, "Compte Lydia enregistré avec succès.")
        return redirect(reverse("add-event"))

    form = LydiaTokenForm(request.POST or None)

    c = {
        "form": form,
    }

    if request.method == "POST" and form.is_valid():
        # API login: récupérer les tokens associés au compte
        url = settings.LYDIA_API_URL + "/auth/login.json"
        data = {"phone": form.cleaned_data.get("phone_number"), "password": form.cleaned_data.get("password")}

        response = requests.post(url, data)
        if response.status_code == 200:
            content = loads(response.content)

            if content.get("status") == "ok":
                business_list = content.get("business_list")
                form_list = []
                if business_list:
                    for business in business_list:
                        form_data = {
                            "vendor_token": business.get("api_token"),
                            "private_token": business.get("api_token_id"),
                            "description": business.get("name"),
                        }
                        form_list.append(LydiaAccountForm(form_data))

                else:
                    messages.error(request, "Aucun business associé à ce compte.")

                c.update({"form_list": form_list})

            else:
                messages.error(request, f"Erreur : {content.get('message')}")

        else:
            messages.error(request, f"Erreur lors l'envoie de la requête : code {response.status_code}")

    return render(request, "guichet/lydia_get_token.html", c)


@login_required
def lydia_init_payment(request, registration_id):
    if request.method == "POST":
        form = LydiaPaymentForm(request.POST)
        if form.is_valid():
            phone_number = form.cleaned_data.get("phone_number")
            registration = get_object_or_404(Registration, id=registration_id)
            ticket = registration.ticket
            event = ticket.event
            lydia_account = event.lydia_account

            # Annule la demande de paiement si elle existe déjà
            if registration.lydia_request_id:
                url = settings.LYDIA_API_URL + "/request/cancel.json"
                data = {
                    "vendor_token": lydia_account.vendor_token,
                    "request_id": registration.lydia_request_id,
                }
                try :
                    response = requests.post(url, data, timeout=settings.REQUEST_TIMEOUT)
                except requests.exceptions.ConnectionError:
                    messages.error(
                        request, "Atention, nous n'avons pas pu annuler la demande de paiement précédente. Veuillez réessayer ultérieurement."
                    )

            # API do: envoyer une demande de paiement
            url = settings.LYDIA_API_URL + "/request/do.json"
            data = {
                "message": "Billetterie : " + event.name,
                "amount": registration.get_real_amount(),
                "currency": "EUR",
                "type": "phone",
                "recipient": phone_number,
                "vendor_token": lydia_account.vendor_token,
                'expire_time': 300,
                "confirm_url": settings.BASE_URL + reverse("lydia-webhook-confirm"),
            }

            try:
                response = requests.post(url, data, timeout=settings.REQUEST_TIMEOUT)
            except requests.exceptions.ConnectionError:
                messages.error(
                    request, "Désolé, Lydia est actuellement indisponible. Veuillez réessayer ultérieurement."
                )
            else:
                if response.status_code == 200:
                    content = loads(response.content)
                    error = content.get("error")
                    if error == "0":
                        registration.lydia_request_uuid = content["request_uuid"]
                        registration.lydia_request_id = content["request_id"]
                        registration.save()

                        return redirect(reverse("show-registration", kwargs={"registration_id": registration_id}))

                    else:
                        messages.error(request, f"Erreur : {content.get('message')}")

                else:
                    messages.error(request, f"Erreur lors l'envoie de la requête : code {response.status_code}")

    else:
        form = LydiaPaymentForm()

    c = {
        "form": form,
    }

    return render(request, "guichet/lydia_init_payment.html", c)


def validate_registration(request, registration):
    registration.paid = True
    registration.payment_method = Registration.PAYMENT_LYDIA
    registration.payment_date = timezone.now()
    registration.info_state = registration.STATE_CAN_ENTER
    registration.save()

    LogEntry.objects.get_or_create(
        user=registration.owner,
        event=registration.ticket.event,
        reason=LogEntry.LOG_LYDIA,
        registration=registration,
        operation=f"Place {registration} payée via Lydia",
    )
    mail_confirm_registration(registration)

    messages.success(request, "Billet payé avec succès.")
    return redirect(reverse("list-self-registrations"))

    
class LydiaStatus(IntEnum):
    INVALID = -1
    WAITING_FOR_PAYMENT = 0
    PAID = 1
    REFUSED = 5
    CANCELED = 6


def lydia_get_status(request, registration_id):
    registration = get_object_or_404(Registration, id=registration_id)
    ticket = registration.ticket
    event = ticket.event
    lydia_account = event.lydia_account

    if registration.paid:
        return "Payé."

    # API state: vérifier que le paiement a été effectué
    url = settings.LYDIA_API_URL + "/request/state.json"
    data = {
        "request_uuid": registration.lydia_request_uuid,
        "vendor_token": lydia_account.vendor_token,
    }

    try:
        response = requests.post(url, data, timeout=settings.REQUEST_TIMEOUT)  # TODO: state cache
        if response.status_code != 200:
            return f"Erreur lors l'envoie de la requête : code {response.status_code}"
    except requests.exceptions.ConnectionError:
        return "Désolé, Lydia est actuellement indisponible, nous ne pouvons pas actualiser le statut du paiement."

    should_log = False
    content = loads(response.content)

    # Vérifier la signature reçue
    received_signature = content.get("signature")
    signature_param = {
        "amount": registration.get_real_amount(),
        "request_uuid": registration.lydia_request_uuid,
    }
    if lydia_account.signature_is_valid(received_signature, signature_param):

        if content["state"] != "error":

            state = int(content["state"])
            if state == LydiaStatus.INVALID:
                status_message = "Requête invalide."
            elif state == LydiaStatus.WAITING_FOR_PAYMENT:
                status_message = "Paiement en attente, veuillez recharger la page après avoir confirmé le paiement."
            elif state == LydiaStatus.PAID:  # Paiement accepté
                validate_registration(request, registration)
                status_message = "Payé."
            elif state == LydiaStatus.REFUSED:
                status_message = "Le paiement a été refusé."
                should_log = True
            elif state == LydiaStatus.CANCELED:
                status_message = "La paiement a été annulé."
                should_log = True
            else:
                status_message = "Erreur inconnue."

        else:
            status_message = f"Erreur lors de l'appel à l'API : code {content['error']}. {content['message']})"

    else:
        status_message = "Erreur : signature invalide"

    if should_log:
        try:
            LogEntry.objects.get_or_create(
                user=registration.owner,
                event=registration.ticket.event,
                reason=LogEntry.LOG_LYDIA_ERROR,
                registration=registration,
                operation=f"Place {registration} : Erreur paiement par Lydia : {status_message}",
            )
        except MultipleObjectsReturned:
            pass

    return status_message


@csrf_exempt
def lydia_webhook_confirm(request):
    if request.method == "POST":
        request_id = request.POST["request_id"]
        vendor_token = request.POST["vendor_token"]
        received_signature = request.POST["sig"]

        registration = get_object_or_404(
            Registration, ticket__event__lydia_account__vendor_token=vendor_token, lydia_request_id=request_id
        )
        lydia_account = registration.ticket.event.lydia_account

        signature_param = {
            "currency": request.POST["currency"],
            "request_id": request_id,
            "amount": request.POST["amount"],
            "signed": request.POST["signed"],
            "transaction_identifier": request.POST["transaction_identifier"],
            "vendor_token": vendor_token,
        }

        #todo: annuler les demandes de paiement qui ont été envoyées lorsque la dernière place a été achetée

        if lydia_account.signature_is_valid(received_signature, signature_param):
            validate_registration(request, registration)
            return HttpResponse("Paiement effectué")

    return HttpResponseForbidden()


@login_required
def event_past_list(request):
    events = []
    for e in Event.objects.filter(Q(end_date__lt=timezone.now()) | Q(archived=True)).all().order_by("-end_date"):
        if e.archived and e.is_manager(request.user):
            # les évènements archivés ne sont visibles que par les managers
            events.append(e)
        elif e.event_confidentiality == Event.EVENT_PUBLIC:
            events.append(e)
        elif e.event_confidentiality == Event.EVENT_LOGIN and request.user.is_authenticated:
            events.append(e)
        elif e.event_confidentiality == Event.EVENT_PRIVE and (
            e.at_least_one_ticket_accessible(request.user) or e.has_customer(request.user)
        ):
            events.append(e)
        elif e.event_confidentiality == Event.EVENT_HIDDEN and e.is_manager(request.user):
            events.append(e)
    c = {
        "events": events,
    }

    return render(request, "guichet/event_past_list.html", c)


def event_list(request):
    events = []
    for e in Event.objects.filter(end_date__gt=timezone.now(), archived=False).all():
        if e.event_confidentiality == Event.EVENT_PUBLIC:
            events.append(e)
        elif e.event_confidentiality == Event.EVENT_LOGIN and request.user.is_authenticated:
            events.append(e)
        elif (
            request.user.is_authenticated
            and e.event_confidentiality == Event.EVENT_PRIVE
            and (e.at_least_one_ticket_accessible(request.user) or e.has_customer(request.user))
        ):
            events.append(e)
        elif e.event_confidentiality == Event.EVENT_HIDDEN and e.is_manager(request.user):
            events.append(e)
    c = {
        "events": events,
    }
    return render(request, "guichet/event_list.html", c)


@login_required
def event_edit(request, event_id=None):
    if event_id:  # modification d’un évènement existant
        event = Event.get_event_for_request(request, event_id)
        if not request.event.is_manager(request.user):
            raise PermissionDenied
    else:  # création d’un évènement
        event = None
        if not request.user.has_perm("guichet.add_event"):
            raise PermissionDenied

    form = EventForm(request.POST or None, instance=event)

    if request.method == "POST" and form.is_valid():

        modification = event

        event = form.save()

        if modification:
            messages.success(request, "Évènement modifié avec succès.")
            if "paypal_account" in form.changed_data:
                if event.paypal_account.no_account:
                    LogEntry(
                        user=request.user,
                        event=event,
                        reason=LogEntry.LOG_EVENT_MAJOR_CHANGE,
                        operation="Désactivation du compte PayPal",
                    ).save()
                    for ticket in event.tickets.all():
                        ticket.online_payment = False
                        ticket.save()
                else:
                    LogEntry(
                        user=request.user,
                        event=event,
                        reason=LogEntry.LOG_EVENT_MAJOR_CHANGE,
                        operation="Changement du compte PayPal pour {}".format(event.paypal_account.email),
                    ).save()
        else:
            event.save()
            UserManager(event=event, user=request.user, access=UserManager.FULL_MANAGER).save()
            messages.success(request, "Évènement créé avec succès.")

        return redirect(reverse("show-event", args=[event.short_url]))

    c = {
        "event": event,
        "form": form,
    }

    return render(request, "guichet/event_edit.html", c)


def event_details(request, event_id):
    event = Event.get_event_for_request(request, event_id, archived=None)
    if event.archived:
        return redirect(reverse("archived-event", args=[event.short_url]))

    if event.event_confidentiality in [Event.EVENT_PUBLIC, Event.EVENT_HIDDEN]:
        all_tickets = event.tickets.order_by("price").all()
        tickets = []
        registrations = 0
        if request.user.is_authenticated:
            for ticket in all_tickets:
                if ticket.is_accessible_for(request.user):
                    tickets += [ticket]
                    registrations += Registration.objects.filter(ticket=ticket, owner=request.user).count()
    else:
        if not request.user.is_authenticated:
            return redirect("{}?next={}".format(settings.LOGIN_URL, request.path))
        if not (event.at_least_one_ticket_accessible(request.user) or event.has_customer(request.user)) and (
            event.event_confidentiality == Event.EVENT_PRIVE
        ):
            raise Http404

        all_tickets = event.tickets.order_by("price").all()
        tickets = []
        registrations = 0
        for ticket in all_tickets:
            if ticket.is_accessible_for(request.user):
                tickets += [ticket]
                registrations += Registration.objects.filter(ticket=ticket, owner=request.user).count()

    c = {
        "tickets": tickets,
        "registrations": registrations,
    }

    return render(request, "guichet/event_details.html", c)


@manager_required
def event_archived(request, event_id):
    Event.get_event_for_request(request, event_id, archived=True)

    return render(request, "guichet/event_details_archived.html", {})


@manager_protected(Manager.EVENT_MANAGER)
@require_http_methods(["POST"])
def event_delete(request, event_id):
    event = Event.get_event_for_request(request, event_id)
    registrations = Registration.objects.filter(ticket__event=event, paid=True)
    if registrations.exists():
        messages.error(request, "Des places ont déjà été achetées pour cet évènement.")
        return redirect(reverse("show-event", args=[event.short_url]))

    event.delete()

    messages.success(request, "Évènement supprimé avec succès.")

    return redirect(reverse("list-events"))


@manager_required
def ticket_list(request, event_id):
    event = Event.get_event_for_request(request, event_id)
    tickets = Ticket.objects.filter(event=event)

    info_sup_before = {}
    info_sup_after = {}
    options = {}
    options_exlusive = {}

    for ticket in tickets:
        q = ticket.options.all()
        options[ticket] = q.filter(group=None)

        # Gestion des groupes d'options mutuellement exclusives
        optionsWithGroup = q.exclude(group=None).order_by("group")
        number_options_list = []
        for o in optionsWithGroup:
            if o.group not in number_options_list:
                number_options_list.append(o.group)
        options_exlusive[ticket] = [(q.filter(group=i), q.filter(group=i).first()) for i in number_options_list]
        # fin des groupes d'options mutuellement exclusifs

        info_sup_before[ticket] = Info_Sup.objects.filter(ticket=ticket, after_payement=False)
        info_sup_after[ticket] = Info_Sup.objects.filter(ticket=ticket, after_payement=True)

    c = {
        "tickets": tickets,
        "options": options,
        "options_exlusive": options_exlusive,
        "info_sup_before": info_sup_before,
        "info_sup_after": info_sup_after,
        "display_sandbox_ticket": event.start_date > timezone.now(),
    }

    return render(request, "guichet/admin_tickets.html", c)


@manager_required
def ticket_details(request, event_id, ticket_id):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)

    q = ticket.options.all()
    options = q.filter(group=None)

    # Gestion des groupes d'options mutuellement exclusives
    optionsWithGroup = q.exclude(group=None).order_by("group")
    number_options_list = []
    for o in optionsWithGroup:
        if o.group not in number_options_list:
            number_options_list.append(o.group)
    options_exlusive = [(q.filter(group=i), q.filter(group=i).first(), i) for i in number_options_list]
    # fin des groupes d'options mutuellement exclusifs

    info_sup_before = Info_Sup.objects.filter(ticket=ticket, after_payement=False)
    info_sup_after = Info_Sup.objects.filter(ticket=ticket, after_payement=True)

    c = {
        "ticket": ticket,
        "options": options,
        "options_exlusive": options_exlusive,
        "info_sup_before": info_sup_before,
        "info_sup_after": info_sup_after,
        "editable": not Registration.objects.filter(ticket=ticket).exists(),
        "display_sandbox_ticket": event.start_date > timezone.now(),
    }

    return render(request, "guichet/admin_tickets_details.html", c)


@manager_required
def ticket_template(request, event_id, ticket_id):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)

    if not ticket.design:
        raise Http404

    path = os.path.join(settings.MEDIA_ROOT, ticket.design.path)

    return sendfile(request, path)


@manager_protected(Manager.EVENT_MANAGER)
def ticket_add(request, event_id, ticket_type=None):
    event = Event.get_event_for_request(request, event_id)
    if not ticket_type:
        return render(request, "guichet/ticket_add.html")
    online_payment_impossible = event.paypal_account.no_account and event.lydia_account.no_account
    froze_field = ["online_payment"] if online_payment_impossible else []
    default_value = {"online_payment": False} if online_payment_impossible else None

    if ticket_type == "open":
        form = OpenTicketForm(
            request.POST or None, request.FILES or None, initial=default_value, forzeField=froze_field
        )
    else:
        assert ticket_type == "restricted"
        form = RestrictedTicketForm(
            request.POST or None, request.FILES or None, initial=default_value, forzeField=froze_field
        )

    if request.method == "POST" and form.is_valid():

        ticket = form.save(commit=False)
        ticket.id = Ticket.next_id(event)
        ticket.event = event
        if ticket_type == "open":
            ticket.allowed_external = True
        ticket.save()
        form.save_m2m()

        messages.success(request, "Billet créé avec succès.")
        return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))

    return render(
        request,
        "guichet/ticket_edit.html",
        {
            "form": form,
        },
    )


@manager_protected(Manager.EVENT_MANAGER)
def ticket_edit(request, event_id, ticket_id):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)

    frozeField = ["price"] if Registration.objects.filter(ticket=ticket, paid=True).exists() else []
    if event.paypal_account.no_account:
        frozeField += ["online_payment"]

    if ticket.allowed_external:
        form = OpenTicketForm(request.POST or None, request.FILES or None, instance=ticket, forzeField=frozeField)
    else:
        form = RestrictedTicketForm(request.POST or None, request.FILES or None, instance=ticket, forzeField=frozeField)

    if request.method == "POST" and form.is_valid():

        ticket = form.save()

        if form.cleaned_data["max_purchase_date"] and form.cleaned_data["max_purchase_date"] > event.end_date:
            ticket.max_purchase_date = None
            ticket.save()

        messages.success(request, "Billet modifié avec succès.")
        return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))

    return render(
        request,
        "guichet/ticket_edit.html",
        {
            "form": form,
            "ticket": ticket,
        },
    )


@manager_protected(Manager.EVENT_MANAGER)
def infosup_edit(request, event_id, ticket_id, infos_id=None, infos_type=None):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)
    restrict = Registration.objects.filter(ticket=ticket).exists()

    if restrict and not infos_id:
        messages.error(request, "Impossible d'ajouter des infos suplémentaires quand il y a des (pré)réservations")
        return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))

    if infos_id is None and infos_type is None:
        c = {
            "event": event,
            "ticket": ticket,
            "Info_Sup": Info_Sup,
        }
        return render(request, "guichet/ticket_choice_infosup.html", c)

    if infos_id:
        infos_sup = get_object_or_404(Info_Sup, ticket=ticket, id=infos_id).get_sup_type()
        info_type = infos_sup.type_info
    else:
        infos_sup = None
        info_type = int(infos_type)

    type_form = InfoSupCreationForm.get_info_form(info_type)
    restrict_field = InfoSupCreationForm.get_frozen_field(info_type)
    choice_form = None

    if infos_id:
        if restrict:
            form = type_form(request.POST or None, instance=infos_sup, forzeField=restrict_field)
            if info_type == Info_Sup.INFO_CHOICE:
                choice_form = InfoSupChoicesSetEditFormSet(request.POST or None, instance=infos_sup)
        else:
            form = type_form(request.POST or None, instance=infos_sup)
            if info_type == Info_Sup.INFO_CHOICE:
                choice_form = InfoSupChoicesSetCreationFormSet(request.POST or None, instance=infos_sup)
    else:
        form = type_form(request.POST or None)
        if info_type == Info_Sup.INFO_CHOICE:
            choice_form = InfoSupChoicesSetCreationFormSet(request.POST or None)

    if request.method == "POST" and form.is_valid():
        infos_sup_recv = form.save(commit=False)
        if choice_form:
            if restrict:
                choice_form = InfoSupChoicesSetEditFormSet(request.POST, instance=infos_sup_recv)
            else:
                choice_form = InfoSupChoicesSetCreationFormSet(request.POST, instance=infos_sup_recv)
        if choice_form is None or choice_form.is_valid():
            infos_sup_recv.ticket = ticket
            infos_sup_recv.type_info = info_type
            if infos_sup and infos_sup_recv.info_validate and infos_sup_recv.type_info != Info_Sup.INFO_SIMPLE:
                if not Info_Sup.objects.get(ticket=ticket, id=infos_id).info_validate:
                    LogEntry(
                        user=request.user,
                        event=event,
                        reason=LogEntry.LOG_EVENT_MAJOR_CHANGE,
                        operation="Activation de la validation automatique "
                        "sur l'information {} ({}) du billet {} ({})".format(
                            infos_sup.name, infos_sup.pk, infos_sup.ticket.name, infos_sup.ticket.id
                        ),
                    ).save()
                    infos_sup_recv.received.filter(state=Info_Sup_Recv.STATE_PENDING).update(
                        state=Info_Sup_Recv.STATE_ACCEPTED
                    )
            infos_sup_recv.save()
            if choice_form:
                choice_form.save()

            messages.success(
                request,
                "Infos {} modifié".format(infos_sup_recv.name)
                if infos_id
                else "Infos {} créée".format(infos_sup_recv.name),
            )
            return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))

    c = {
        "form": form,
        "infosup": infos_sup,
        "event": event,
        "ticket": ticket,
        "desc": Info_Sup.INFO_DESC[info_type],
        "choice_form": choice_form,
    }

    return render(request, "guichet/ticket_edit_infosup.html", c)


@manager_protected(Manager.EVENT_MANAGER)
@require_http_methods(["POST"])
def infosup_del(request, event_id, ticket_id, infos_id):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)

    if Registration.objects.filter(ticket=ticket).exists():
        messages.error(request, "Impossible de supprimer des infos suplémentaires quand il y a des (pré)réservations")
        return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))

    infos_sup = get_object_or_404(Info_Sup, ticket=ticket, id=infos_id)
    infos_sup.delete()
    messages.success(request, "Informations supprimées")
    return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))


@manager_protected(Manager.EVENT_MANAGER)
def option_alone_edit(request, event_id, ticket_id, option_id=None):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)

    if Registration.objects.filter(ticket=ticket).exists() and not option_id:
        messages.error(request, "Impossible d'ajouter des options quand il y a des (pré)réservations")
        return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))

    if option_id:
        option = get_object_or_404(Option, ticket=ticket, id=option_id, group=None)
        if Registration.objects.filter(ticket=ticket).exists():

            form = OptionAloneCreationForm(request.POST or None, instance=option, forzeField=["price", "admin_only"])
        else:
            form = OptionAloneCreationForm(request.POST or None, instance=option)
    else:
        option = None
        form = OptionAloneCreationForm(request.POST or None)

    if request.method == "POST" and form.is_valid():
        option = form.save(commit=False)
        option.ticket = ticket
        option.save()
        messages.success(
            request, "Option {} modifiée".format(option.name) if option_id else "Option {} créée".format(option.name)
        )
        return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))

    c = {
        "form": form,
        "option": option,
        "event": event,
        "ticket": ticket,
    }

    return render(request, "guichet/ticket_edit_option_alone.html", c)


@manager_protected(Manager.EVENT_MANAGER)
@require_http_methods(["POST"])
def option_alone_del(request, event_id, ticket_id, option_id):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)

    if Registration.objects.filter(ticket=ticket).exists():
        messages.error(request, "Impossible de supprimer des options quand il y a des (pré)réservations")
        return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))

    option = get_object_or_404(Option, ticket=ticket, id=option_id, group=None)
    option.delete()
    messages.success(request, "Option supprimée")
    return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))


@manager_protected(Manager.EVENT_MANAGER)
def option_group_edit(request, event_id, ticket_id, group_id=None):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)

    restrict = Registration.objects.filter(ticket=ticket).exists()
    if restrict and not group_id:
        messages.error(request, "Impossible d'ajouter des options quand il y a des (pré)réservations")
        return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))

    if group_id:
        options = Option.objects.filter(ticket=ticket, group=group_id)
        if not options.exists():
            raise Http404
        inst = {
            "group_name": options.first().group_name,
            "description": options.first().description,
            "mandatory": options.first().mandatory,
        }
        form = OptionGroupCreationForm(request.POST if request.method == "POST" else inst)
        if restrict:
            formoptions = OptionGroupSetEditFormSet(
                request.POST or None, queryset=options, form_kwargs={"forzeField": ["price", "admin_only"]}
            )
            form.fields["mandatory"].disabled = True
        elif request.managers_access < Manager.EVENT_MANAGER:
            formoptions = OptionGroupSetEditFormSet(request.POST or None, queryset=options)
        else:
            formoptions = OptionGroupSetCreationFormSet(request.POST or None, queryset=options)

    else:
        options = None
        form = OptionGroupCreationForm(request.POST or None)
        formoptions = OptionGroupSetCreationFormSet(request.POST or None, queryset=Option.objects.none())

    if request.method == "POST" and form.is_valid() and formoptions.is_valid():
        group_id = group_id or Option.get_new_ID_group(ticket)
        options_out = formoptions.save(commit=False)
        if restrict:
            options_out = [obj[0] for obj in formoptions.changed_objects]
        else:
            # Suppression des élements
            for option in formoptions.deleted_objects:
                option.delete()
        # enregistrement des modifications sur les objects modifiés ou créés
        for option in options_out:
            option.ticket = ticket
            option.group_name = form.cleaned_data["group_name"]
            option.description = form.cleaned_data["description"]
            option.mandatory = form.cleaned_data["mandatory"]
            option.group = group_id
            option.save()
        # enresitrement des changements sur toutes les options (peut être redondant)
        for option in Option.objects.filter(ticket=ticket, group=group_id):
            option.group_name = form.cleaned_data["group_name"]
            option.description = form.cleaned_data["description"]
            option.mandatory = form.cleaned_data["mandatory"]
            option.save()

        if (
            form.cleaned_data["mandatory"]
            and not Option.objects.filter(ticket=ticket, group=group_id, price=0, admin_only=False).exists()
        ):
            # si une option est obligatoire et qu'il n'y a pas de choix disponible gratuit, on en ajoute un
            Option(
                ticket=ticket,
                group=group_id,
                group_name=form.cleaned_data["group_name"],
                description=form.cleaned_data["description"],
                mandatory=True,
                price=0,
                admin_only=False,
                name="Sans l'option",
            ).save()
            messages.warning(request, "Une option gratuite et disponible a été ajoutée.")

        messages.success(
            request,
            "Options groupées {} modifiées".format(form.cleaned_data["group_name"])
            if group_id
            else "Options groupées {} créées".format(form.cleaned_data["group_name"]),
        )
        return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))

    c = {
        "form": form,
        "formoptions": formoptions,
        "options": options,
        "event": event,
        "ticket": ticket,
        "restrict": restrict,
    }

    return render(request, "guichet/ticket_edit_option_group.html", c)


@manager_protected(Manager.EVENT_MANAGER)
@require_http_methods(["POST"])
def option_group_del(request, event_id, ticket_id, group_id):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)

    if Registration.objects.filter(ticket=ticket).exists():
        messages.error(request, "Impossible de supprimer des options quand il y a des (pré)réservations")
        return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))

    options = Option.objects.filter(ticket=ticket, group=group_id)
    if not options.exists():
        raise Http404
    for opt in options:
        opt.delete()
    messages.success(request, "Groupe d'option supprimé")
    return redirect(reverse("details-ticket", args=[event.short_url, ticket.id]))


@manager_protected(Manager.EVENT_MANAGER)
@require_http_methods(["POST"])
def ticket_delete(request, event_id, ticket_id):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)

    registrations = Registration.objects.filter(ticket=ticket, paid=True)

    if registrations.exists():
        messages.error(request, "Des places ont déjà été achetées pour ce billet.")
    else:
        ticket.delete()
        messages.success(request, "Billet supprimé avec succès.")

    return redirect(reverse("list-tickets", args=[event.short_url]))


@manager_required
def ticket_test_template(request, event_id, ticket_id):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)

    if ticket.event.start_date < timezone.now():
        messages.error(request, "Impossible de tester un billet après la date de début de l'évènement !")
        return redirect(reverse("list-tickets", args=[event.short_url]))
    elif not ticket.design:
        messages.error(request, "Pas de design pour ce billet.")
        return redirect(reverse("list-tickets", args=[event.short_url]))
    else:
        img = print_sandbox_billet(ticket, "PDF")

        response = HttpResponse(img, content_type="application/x-pdf")
        response["Content-Disposition"] = 'attachment; filename="test-template.pdf"'
        return response


@login_required
def ticket_purchase(request, event_id, ticket_id):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)

    # Ce billet n’est pas accessible à cet utilisateur
    if not ticket.is_accessible_for(request.user):
        messages.error(request, "Ce billet ne vous est pas accessible.")
        return redirect(reverse("show-event", args=[event.short_url]))

    payee = request.GET.get("payee", None)
    user_have_registration = ticket.event.registrations.filter(owner=request.user, payee__isnull=True).exists()
    user_registration_id = None
    if user_have_registration:
        user_registration_id = ticket.event.registrations.filter(owner=request.user, payee__isnull=True)[0].id

    if payee and payee not in ["oneself", "godson"]:
        raise Http404

    # L’utilisateur essaye d’acheter une deuxième place pour lui même
    if payee == "oneself" and user_have_registration:
        messages.error(request, "Vous possédez déjà une place pour cet évènement.")
        return redirect(reverse("list-self-registrations"))

    # Plus de place disponible pour cet évènement
    if not ticket.is_remain():
        messages.error(request, "Toutes les places disponibles ont été vendues.")
        return redirect(reverse("show-event", args=[event.short_url]))

    if not ticket.still_available():
        messages.error(request, "Ce billet n’est plus ou pas encore disponible à la vente.")
        return redirect(reverse("show-event", args=[event.short_url]))

    godson_count = ticket.registrations.filter(owner=request.user, payee__isnull=False).count()
    godson_remaining = event.limit_godson - godson_count
    # Billet non extérieur : on vérifie les contraintes de parrainage
    if payee == "godson" and not ticket.allowed_external:
        # Parrainage non autorisé
        if not ticket.allowed_godson:
            messages.error(request, "Ce billet n’est pas accessible aux extérieurs, y compris avec parrainage.")
            return redirect(reverse("show-event", args=[event.short_url]))
        # Quotas de parrainage dépassé
        elif godson_remaining <= 0:
            messages.error(request, "Vous ne pouvez pas parrainer plus de {} personne(s).".format(event.limit_godson))
            return redirect(reverse("show-event", args=[event.short_url]))

    form = GodsonForm(request.POST or None)
    term_of_use_form = TermOfUseForm(request.POST or None)
    q = ticket.options.all().filter(admin_only=False)
    options = q.filter(group=None)

    # Gestion des groupes d'options mutuellement exclusives
    optionsWithGroup = q.exclude(group=None).order_by("group")
    numbers_options_list = []
    for o in optionsWithGroup:
        if o.group not in numbers_options_list:
            numbers_options_list.append(o.group)
    options_form = ChoiceOptionsForm(request.POST or None, options=options)
    q_list = [(q.filter(group=i), i) for i in numbers_options_list]
    options_with_group_form_list = [
        ChoiceOptionsWithGroupForm(request.POST or None, queryset=q[0], index=q[1]) for q in q_list
    ]
    # fin des groupes d'options mutuellement exclusifs

    info_sup = Info_Sup.objects.filter(ticket=ticket)

    conditions_field = None
    if request.method == "POST":
        if term_of_use_form.is_valid():
            conditions_field = term_of_use_form.cleaned_data["conditions_field"]
        # Si la case n'est pas cochée
        if not conditions_field:
            messages.error(
                request, "Veuillez accepter la charte de la billetterie et les conditions de vente de cet évènement."
            )
    if request.method == "POST" and payee and (conditions_field or not event.terms_of_use):
        is_valid_list = [f.is_valid() for f in options_with_group_form_list]
        if options_form.is_valid() and all(is_valid_list) and (payee == "oneself" or form.is_valid()):
            # parainnage
            if payee == "godson":
                registration = Registration(
                    owner=request.user,
                    ticket=ticket,
                    paid=False,
                    payee=form.cleaned_data["payee"],
                    conditions_accepted=conditions_field,
                )
            # billet perso
            else:
                registration = Registration(
                    owner=request.user, ticket=ticket, paid=False, conditions_accepted=conditions_field
                )
                if conditions_field and (
                    request.user.charte_date is None or request.user.charte_version < settings.CHARTE_VERSION
                ):
                    request.user.charte_date = timezone.now()
                    request.user.charte_version = settings.CHARTE_VERSION
                    request.user.save()

            registration.save()
            for id in options_form.cleaned_data["options"]:
                o = get_object_or_404(Option, pk=id)
                registration.options.add(o)
            # Traitement des options mutuellement exclusives
            # On reverse notre liste pour obtenir les champs dans l'ordre
            numbers_options_list.reverse()
            for f in options_with_group_form_list:
                try:
                    id = f.cleaned_data["choice-{}".format(numbers_options_list.pop())].id
                    o = get_object_or_404(Option, pk=id)
                    registration.options.add(o)
                except AttributeError:
                    pass
            registration.save()
            # gestion des billets gratuits
            maj_state_registration(registration)
            if registration.get_real_amount() == 0 and registration.info_state != Registration.STATE_CANNOT_PAID:
                registration.paid = True
                registration.payment_method = Registration.PAYMENT_UNKNOW
                registration.payment_date = timezone.now()
                registration.save()
                mail_confirm_registration(registration)
            return redirect(reverse("show-registration", args=[registration.id]))

    c = {
        "ticket": ticket,
        "godson_limit": event.limit_godson,
        "godson_count": godson_count,
        "godson_remaining": godson_remaining,
        "user_have_registration": user_have_registration,
        "user_registration_id": user_registration_id,
        "form": form,
        "term_of_use_form": term_of_use_form,
        "terms_of_use": event.terms_of_use,
        "options_form": options_form,
        "options_exclusive_form": options_with_group_form_list,
        "info_sup": info_sup,
    }

    return render(request, "guichet/ticket_purchase.html", c)


@manager_required
def stats(request, event_id):
    event = Event.get_event_for_request(request, event_id, archived=None)
    c = get_event_stats(event)

    if event.archived:
        if "version" in c:
            if c["version"] == 3:
                # version 3
                return render(request, "guichet/stats/archived_v3.html", c)
            elif c["version"] == 2:
                # version 2
                return render(request, "guichet/stats/archived_v2.html", c)
            else:
                # version 1
                return render(request, "guichet/stats/archived_v1.html", c)
        else:
            # version 1
            return render(request, "guichet/stats/archived_v1.html", c)
    else:
        return render(request, "guichet/admin_stats.html", c)


@manager_required
def logs_list(request, event_id):
    event = Event.get_event_for_request(request, event_id)
    logs = LogEntry.objects.filter(event=event).order_by("-date").all()
    page = request.GET.get("page")
    paginator = Paginator(logs, settings.ITEMS_PER_PAGE)
    try:
        logs = paginator.page(page)
    except PageNotAnInteger:
        logs = paginator.page(1)
    except EmptyPage:
        logs = paginator.page(paginator.num_pages)

    c = {
        "logs": logs,
    }

    return render(request, "guichet/admin_logs.html", c)


@manager_protected(Manager.FULL_MANAGER)
def manager_list(request, event_id, user_id=None, group_id=None):
    event = Event.get_event_for_request(request, event_id, archived=None)
    userform = UserManagerForm(request.POST or None)
    groupform = GroupManagerForm(request.POST or None)

    if request.method == "GET":
        if user_id:
            user_manager = get_object_or_404(UserManager, pk=user_id, event=event)
            if user_manager.user.login_inp_net:
                userform = UserManagerForm(instance=user_manager, initial={"login": user_manager.user.login_inp_net})
            else:
                userform = UserManagerForm(instance=user_manager, initial={"email": user_manager.user.email})
        elif group_id:
            group_manager = get_object_or_404(GroupManager, pk=group_id, event=event)
            groupform = GroupManagerForm(instance=group_manager)

    elif request.method == "POST":

        if "promote-group" in request.POST and groupform.is_valid():
            group_manager = groupform.save(commit=False)
            old_manager = GroupManager.objects.filter(
                event=event, group=group_manager.group, bureau_only=group_manager.bureau_only
            )
            if old_manager.exists():
                old_manager = old_manager.first()
                if old_manager.access == group_manager.access:
                    messages.warning(request, "Ce groupe possède déjà ces permissions sur l'évènement")
                else:
                    old_manager.access = group_manager.access
                    old_manager.save()
                    if not event.archived:
                        LogEntry(
                            user=request.user,
                            event=event,
                            reason=LogEntry.LOG_MANAGER,
                            operation="Le {} {} a reçu les droits ({})".format(
                                "bureau de" if group_manager.bureau_only else "groupe",
                                group_manager.group,
                                group_manager.get_access_display(),
                            ),
                        ).save()
                    messages.success(request, "Modification des droits effectuée avec succès")
            else:
                group_manager.event = event
                group_manager.save()
                if not event.archived:
                    LogEntry(
                        user=request.user,
                        event=event,
                        reason=LogEntry.LOG_MANAGER,
                        operation="Le {} {} a été promu administrateur avec les droits ({})".format(
                            "bureau de" if group_manager.bureau_only else "groupe",
                            group_manager.group,
                            group_manager.get_access_display(),
                        ),
                    ).save()
                messages.success(
                    request,
                    "{} promu au rang d’administrateur avec succès.".format(
                        "Bureau" if group_manager.bureau_only else "Groupe"
                    ),
                )
            return redirect(reverse("list-managers", args=[event.short_url]))

        elif "promote-user" in request.POST and userform.is_valid():
            manager = None
            if userform.cleaned_data["login"] != "" and userform.cleaned_data["email"] == "":
                manager, message = get_ldap_user(userform.cleaned_data["login"])
                message(request)

            elif userform.cleaned_data["login"] == "" and userform.cleaned_data["email"] != "":
                managers = User.objects.filter(email=userform.cleaned_data["email"], is_anonyme=False, is_active=True)
                if managers.exists():
                    manager = managers.first()
                else:
                    messages.error(request, "Pas de compte associé à cette adresse")

            if manager:
                user_manager = UserManager.objects.filter(event=event, user=manager)
                if user_manager.exists():
                    user_manager = user_manager.first()
                    if userform.cleaned_data["access"] == user_manager.access:
                        messages.warning(request, "Cet utilisateur possède déjà ces permissions sur l'évènement")
                    else:
                        user_manager.access = userform.cleaned_data["access"]
                        user_manager.save()
                        if not event.archived:
                            LogEntry(
                                user=request.user,
                                event=event,
                                reason=LogEntry.LOG_MANAGER,
                                operation="{} a reçu les droits ({})".format(
                                    manager.get_full_name(), user_manager.get_access_display()
                                ),
                            ).save()
                        messages.success(request, "Modification des droits effectuée avec succès")
                else:
                    user_manager = userform.save(commit=False)
                    user_manager.event = event
                    user_manager.user = manager
                    user_manager.save()
                    if not event.archived:
                        LogEntry(
                            user=request.user,
                            event=event,
                            reason=LogEntry.LOG_MANAGER,
                            operation="{} a été promu administrateur avec les droits ({})".format(
                                manager.get_full_name(), user_manager.get_access_display()
                            ),
                        ).save()
                    messages.success(request, "Utilisateur promu au rang d’administrateur avec succès.")
                return redirect(reverse("list-managers", args=[event.short_url]))

    c = {
        "user_managers": UserManager.objects.filter(event=event).order_by("-user__last_login"),
        "group_managers": GroupManager.objects.filter(event=event).order_by("bureau_only", "group__displayName"),
        "userform": userform,
        "groupform": groupform,
    }

    return render(request, "guichet/admin_managers.html", c)


@manager_protected(Manager.FULL_MANAGER)
@require_http_methods(["POST"])
def manager_revoke(request, event_id, user_id=None, group_id=None):
    event = Event.get_event_for_request(request, event_id, archived=None)
    if user_id and group_id:
        raise Http404
    elif user_id:
        user_manager = get_object_or_404(UserManager, event=event, pk=user_id)
        message = "{} a été déchu de son poste d'administrateur ({})".format(
            user_manager.user.get_full_name(), user_manager.get_access_display()
        )
        user_manager.delete()
        messages.success(request, "Utilisateur déchu avec succès.")
    elif group_id:
        group_manager = get_object_or_404(GroupManager, event=event, pk=group_id)
        if group_manager.bureau_only:
            message = "Le bureau de {} a été déchu de son poste d'administrateur ({})".format(
                group_manager.group, group_manager.get_access_display()
            )
            messages.success(request, "Bureau déchu avec succès.")
        else:
            message = "Le Groupe {} a été déchu de son poste d'administrateur ({})".format(
                group_manager.group, group_manager.get_access_display()
            )
            messages.success(request, "Groupe déchu avec succès.")
        group_manager.delete()
    else:
        raise Http404

    if not event.archived:
        LogEntry(user=request.user, event=event, reason=LogEntry.LOG_MANAGER, operation=message).save()

    return redirect(reverse("list-managers", args=[event.short_url]))


@manager_protected(Manager.ENTRY_MANAGER)
def registration_check_with_uuid(request, event_id, registration_uuid):
    event = Event.get_event_for_request(request, event_id)
    registrations = Registration.objects.filter(
        ticket__event=event,
        qrcode_uuid=registration_uuid,
        paid=True,
        opposed=False,
        info_state=Registration.STATE_CAN_ENTER,
    )
    message = None
    registration = None
    expected = None
    entered = None
    total = None
    owner = "Inconnu"

    if len(registrations) == 0:
        if len(Registration.objects.filter(ticket__event=event, qrcode_uuid=registration_uuid, opposed=True)) != 0:
            message = "Ce billet a été mis en opposition !"
        elif len(Registration.objects.filter(ticket__event=event, qrcode_uuid=registration_uuid, paid=False)) != 0:
            message= "Ce billet n’a pas encore été payé !"
        elif len(Registration.objects.filter(ticket__event=event, qrcode_uuid=registration_uuid)) != 0:
            message = "Ce billet est dans un état invalide !"
        else:
            message = "Ce billet n’existe pas."
    elif len(registrations) > 1:
        LogEntry(
            user=request.user,
            event=event,
            reason=LogEntry.LOG_ENTRANCE,
            operation="Erreur: {} a validé un billet lié à plusieurs enregistrements: \
                registrations='{}'".format(request.user.get_full_name(), registrations),
        ).save()
        message = "Erreur critique, plusieurs objets trouvés. Merci de contacter INP-net !"
    if len(registrations) == 1:
        registration = registrations[0]
        owner = registration.owner.first_name + " " + registration.owner.last_name
        if registration.payee:
            owner = owner + " pour " + registration.payee

        if request.method == "POST":
            if registration.paid and not registration.opposed and not registration.used and registration.info_state == Registration.STATE_CAN_ENTER:
                message = "Billet valide."
                registration.used = True
                registration.used_date = timezone.now()
                registration.save()
                LogEntry(
                    user=request.user,
                    event=event,
                    reason=LogEntry.LOG_ENTRANCE,
                    registration=registration,
                    operation=f"""{{registration}} est entré au poste tenu par {request.user.get_full_name()}""",
                ).save()
            elif registration.paid and not registration.opposed and registration.used and registration.info_state == Registration.STATE_CAN_ENTER:
                message = "Billet déjà validé" + timezone.localtime(registration.used_date).strftime(" le %d/%m/%Y à %H:%M:%S")
    if request.method == "POST":
        data = {
            "ticket": registration.ticket.name,
            "message": message,
            "entered": Registration.objects.filter(ticket__event=event, paid=True, opposed=False, info_state=Registration.STATE_CAN_ENTER).filter(used=True).count(),
            "expected": Registration.objects.filter(ticket__event=event, paid=True, opposed=False, info_state=Registration.STATE_CAN_ENTER).filter(used=False).count(),
            "total": Registration.objects.filter(ticket__event=event, paid=True, opposed=False, info_state=Registration.STATE_CAN_ENTER).count(),
            "owner": owner,
            }
        return JsonResponse(data)
    else:
        data = {
            "registration": registration,
            "tickets": event.tickets.filter(allow_anonyme=True).order_by("price"),
            "message": message,
            "entered": Registration.objects.filter(ticket__event=event, paid=True, opposed=False, info_state=Registration.STATE_CAN_ENTER).filter(used=True).count(),
            "expected": Registration.objects.filter(ticket__event=event, paid=True, opposed=False, info_state=Registration.STATE_CAN_ENTER).filter(used=False).count(),
            "total": Registration.objects.filter(ticket__event=event, paid=True, opposed=False, info_state=Registration.STATE_CAN_ENTER).count(),
            "owner": owner,
            }
        return render(request, "guichet/admin_checkpoint_with_id.html", data)


@manager_protected(Manager.ENTRY_MANAGER)
def registration_check(request, event_id):
    event = Event.get_event_for_request(request, event_id)

    form = CheckForm(request.POST or None)
    valid = None
    message = None
    registration = None
    if request.method == "POST" and form.is_valid():

        search = form.cleaned_data["search"]
        valid, arg = check_registration(event, search)

        if valid:
            registration = arg
            message = "Billet valide."
            registration.used = True
            registration.used_date = timezone.now()
            registration.save()
            LogEntry(
                user=request.user,
                event=event,
                reason=LogEntry.LOG_ENTRANCE,
                registration=registration,
                operation="{{registration}} est entré au poste tenu par {}".format(request.user.get_full_name()),
            ).save()
        else:
            sandbox, argsandbox = check_registration_sandbox(event, search)
            if sandbox:
                arg = "Billet de test (invalide, ne pas laisser entrer) - ({})".format(argsandbox.name)
            message = arg

        form = CheckForm()

    if valid is None:
        cls = "primary"
    elif valid:
        cls = "success"
    else:
        cls = "danger"

    registrations = Registration.objects.filter(
        ticket__event=event, paid=True, opposed=False, info_state=Registration.STATE_CAN_ENTER
    )

    tickets = event.tickets.filter(allow_anonyme=True).order_by("price")

    c = {
        "form": form,
        "cls": cls,
        "message": message,
        "registration": registration,
        "entered": registrations.filter(used=True).count(),
        "expected": registrations.filter(used=False).count(),
        "total": registrations.count(),
        "tickets": tickets,
    }

    return render(request, "guichet/admin_checkpoint.html", c)

@manager_protected(Manager.ENTRY_MANAGER)
def registration_check_with_QR(request, event_id):
    event = Event.get_event_for_request(request, event_id)

    registration = None

    registrations = Registration.objects.filter(
        ticket__event=event, paid=True, opposed=False, info_state=Registration.STATE_CAN_ENTER
    )

    c = {
        "registration": registration,
        "event_id": event.id,
        "entered": registrations.filter(used=True).count(),
        "expected": registrations.filter(used=False).count(),
        "total": registrations.count(),
    }

    return render(request, "guichet/admin_checkpoint_with_qr.html", c)


@manager_required
def registration_list_paid(request, event_id):
    event = Event.get_event_for_request(request, event_id)
    registrations = Registration.objects.filter(ticket__event=event, paid=True).order_by(
        "-payment_date", "-purchase_date"
    )
    registrations_filter = RegistrationPaidFilter(request.GET, event=event, queryset=registrations)
    page = request.GET.get("page")
    paginator = Paginator(registrations_filter.qs, settings.ITEMS_PER_PAGE)
    try:
        pagined_registrations = paginator.page(page)
    except PageNotAnInteger:
        pagined_registrations = paginator.page(1)
    except EmptyPage:
        pagined_registrations = paginator.page(paginator.num_pages)
    emails = (
        registrations.filter(owner__is_anonyme=False)
        .order_by("owner__email")
        .values_list("owner__email", flat=True)
        .distinct()
    )

    c = {
        "registrations": pagined_registrations,
        "filters": registrations_filter,
        "count_without_opposed_and_doc_expected": registrations.filter(
            opposed=False, info_state=Registration.STATE_CAN_ENTER
        ).count(),
        "count": registrations.count(),
        "emails": ",".join(emails),
    }

    return render(request, "guichet/admin_list_paid_registrations.html", c)


@manager_required
def registration_list_unpaid(request, event_id):
    event = Event.get_event_for_request(request, event_id)
    registrations = Registration.objects.filter(ticket__event=event, paid=False).order_by("-purchase_date")
    registrations_filter = RegistrationUnpaidFilter(request.GET, event=event, queryset=registrations)
    page = request.GET.get("page")
    paginator = Paginator(registrations_filter.qs, settings.ITEMS_PER_PAGE)
    try:
        pagined_registrations = paginator.page(page)
    except PageNotAnInteger:
        pagined_registrations = paginator.page(1)
    except EmptyPage:
        pagined_registrations = paginator.page(paginator.num_pages)
    emails = (
        registrations.filter(owner__is_anonyme=False)
        .order_by("owner__email")
        .values_list("owner__email", flat=True)
        .distinct()
    )

    c = {
        "registrations": pagined_registrations,
        "filters": registrations_filter,
        "emails": ",".join(emails),
        "count": registrations.count(),
        "count_doc_expected_only": registrations.filter(info_state=Registration.STATE_CANNOT_PAID).count(),
        "methods": list(filter(lambda m: m[0] in Registration.ALLOWED_ADMIN_PAYMENT, Registration.PAYMENT_METHODS)),
    }

    return render(request, "guichet/admin_list_unpaid_registrations.html", c)


@manager_required
def registration_download(request, event_id):
    event = Event.get_event_for_request(request, event_id)
    response = HttpResponse(content_type="application/pdf")
    response["Content-Disposition"] = 'attachment; filename="registrations.pdf"'

    buffer = BytesIO()

    list_registrations(event, buffer)

    pdf = buffer.getvalue()
    buffer.close()
    response.write(pdf)

    return response


@manager_required
def registration_choice(request, event_id):
    event = Event.get_event_for_request(request, event_id)
    all_tickets = event.tickets.order_by("price").all()

    c = {
        "tickets": all_tickets,
    }

    return render(request, "guichet/admin_choice_registration.html", c)


@manager_protected(Manager.EVENT_MANAGER)
def registration_new(request, event_id, ticket_id):
    event = Event.get_event_for_request(request, event_id)
    ticket = get_object_or_404(Ticket, event=event, id=ticket_id)

    force_remain = "force_remain" in request.GET and not ticket.is_remain()
    if not ticket.is_remain() and not force_remain:
        messages.error(request, "Toutes les places disponibles ont été vendues.")
        return redirect(reverse("choice-registration", args=[event.short_url]))

    force_available = "force_available" in request.GET and not ticket.still_available()
    if not ticket.still_available() and not force_available:
        messages.error(request, "Ce billet n’est plus ou pas encore disponible à la vente.")
        return redirect(reverse("choice-registration", args=[event.short_url]))

    q = ticket.options.all()
    options = q.filter(group=None)

    # Gestion des groupes d'options mutuellement exclusives
    optionsWithGroup = q.exclude(group=None).order_by("group")
    numbers_options_list = []
    for o in optionsWithGroup:
        if o.group not in numbers_options_list:
            numbers_options_list.append(o.group)
    options_form = ChoiceOptionsForm(request.POST or None, options=options)
    q_list = [(q.filter(group=i), i) for i in numbers_options_list]
    options_with_group_form_list = [
        ChoiceOptionsWithGroupForm(request.POST or None, queryset=q[0], index=q[1]) for q in q_list
    ]
    # fin des groupes d'options mutuellement exclusifs

    form = RegistrationForm(request.POST or None)
    display_force_field = False

    if request.method == "POST":
        is_valid_list = [f.is_valid() for f in options_with_group_form_list]
        if form.is_valid() and options_form.is_valid() and all(is_valid_list):
            buyer = None
            anonyme = form.cleaned_data["anonyme"] if ticket.allow_anonyme else False
            payee = form.cleaned_data["payee"] or None

            if sum([bool(form.cleaned_data["login"]), bool(form.cleaned_data["email"]), anonyme]) > 1:
                messages.error(
                    request,
                    "Veuillez renseigner <strong>soit</strong> un login INP-net, "
                    "<strong>soit</strong> une adresse email <strong>ou</strong> "
                    "cocher la cas 'Vente sur Place' si le ticket le permet.",
                )
            elif form.cleaned_data["login"]:
                buyer, message = get_ldap_user(form.cleaned_data["login"])
                message(request)
            elif form.cleaned_data["email"]:
                buyer = User.objects.filter(email=form.cleaned_data["email"], is_anonyme=False, is_active=True)
                if buyer.exists():
                    buyer = buyer.first()
                else:
                    messages.error(request, "Cette adresse email n’a pas été trouvée.")
            elif anonyme:
                buyer = User.get_anonyme_user()
            else:
                messages.error(
                    request,
                    "Veuillez renseigner un login INP-net ou une adresse email "
                    "ou cocher la case 'Vente sur Place' si le ticket le permet.",
                )

            conditions_field = form.cleaned_data["conditions_field"]
            force_condition = form.cleaned_data["force_condition"]
            forced_condition = False
            valid = True
            # check condition billet
            if not buyer:
                valid = False
            if valid and not payee:
                if buyer.is_anonyme:
                    messages.error(request, "Merci de préciser un nom pour le ticket.")
                    valid = False
                elif ticket.event.registrations.filter(owner=buyer, payee__isnull=True).exists():
                    messages.error(request, "Cet utilisateur possède déjà une place pour cet évènement.")
                    valid = False
                elif not ticket.is_accessible_for(buyer, with_admin=True):
                    forced_condition = True
                    if not force_condition:
                        valid = False
                        messages.warning(
                            request,
                            "Cet utilisateur ne peux pas prendre ce billet. "
                            "Vous pourvez néanmoins forcer la validation.",
                        )
            if valid and payee:
                if not ticket.allowed_external and not buyer.is_anonyme:
                    if not ticket.is_accessible_for(buyer, with_admin=True):
                        forced_condition = True
                        if not force_condition:
                            valid = False
                            messages.warning(
                                request,
                                "Cet utilisateur ne peux pas prendre ce billet. "
                                "Vous pourvez néanmoins forcer la validation.",
                            )
                    if not ticket.allowed_godson:
                        forced_condition = True
                        if not force_condition:
                            valid = False
                            messages.warning(
                                request,
                                "Ce billet n’est pas accessible aux extérieurs, "
                                "y compris avec parrainage. Vous pourvez néanmoins forcer la validation.",
                            )
                    elif event.limit_godson <= ticket.registrations.filter(owner=buyer, payee__isnull=False).count():
                        forced_condition = True
                        if not force_condition:
                            valid = False
                            messages.warning(
                                request,
                                "Cet utilisateur est déjà parrain de {} personne(s). "
                                "Vous pourvez néanmoins forcer la validation.".format(event.limit_godson),
                            )

            if not valid:
                display_force_field = forced_condition
            else:
                registration = Registration(
                    owner=buyer, ticket=ticket, paid=False, payee=payee, conditions_accepted=conditions_field
                )
                registration.save()
                if force_available:
                    LogEntry(
                        user=request.user,
                        event=event,
                        registration=registration,
                        reason=LogEntry.LOG_FORCE_LIMIT,
                        operation="Création du billet {registration} en dehors des dates de vente",
                    ).save()
                if force_remain:
                    LogEntry(
                        user=request.user,
                        event=event,
                        registration=registration,
                        reason=LogEntry.LOG_FORCE_LIMIT,
                        operation="Création du billet {registration} en dépassant le nombre de place défini",
                    ).save()
                if forced_condition or (buyer.is_anonyme and not ticket.allowed_external):
                    LogEntry(
                        user=request.user,
                        event=event,
                        registration=registration,
                        reason=LogEntry.LOG_FORCE_LIMIT,
                        operation="Création du billet {registration} en forçant les contraintes",
                    ).save()
                for id in options_form.cleaned_data["options"]:
                    o = get_object_or_404(Option, pk=id)
                    registration.options.add(o)
                # Traitement des options mutuellement exclusives
                # On reverse notre liste pour obtenir les champs dans l'ordre
                numbers_options_list.reverse()
                for f in options_with_group_form_list:
                    try:
                        id = f.cleaned_data["choice-{}".format(numbers_options_list.pop())].id
                        o = get_object_or_404(Option, pk=id)
                        registration.options.add(o)
                    except AttributeError:
                        pass

                registration.save()
                # gestion des billets gratuits
                maj_state_registration(registration)
                if registration.get_real_amount() == 0 and registration.info_state != Registration.STATE_CANNOT_PAID:
                    registration.paid = True
                    registration.payment_method = Registration.PAYMENT_UNKNOW
                    registration.payment_date = timezone.now()
                    registration.save()
                    mail_confirm_registration(registration)
                messages.success(
                    request, "Place préréservée avec succès pour l’utilisateur {}.".format(buyer.get_full_name())
                )
                return redirect(reverse("registration-edit", args=[event.short_url, registration.id]))

    c = {
        "form": form,
        "force_remain": force_remain,
        "force_available": force_available,
        "terms_of_use": event.terms_of_use,
        "options_form": options_form,
        "options_exclusive_form": options_with_group_form_list,
        "ticket": ticket,
        "display_force_field": display_force_field,
        "display_anonyme": ticket.allow_anonyme,
    }

    return render(request, "guichet/admin_new_registration.html", c)


@manager_protected(Manager.EVENT_MANAGER)
@require_http_methods(["POST"])
def registration_validate(request, event_id, registration_id, payment_method=Registration.PAYMENT_UNKNOW):
    event = Event.get_event_for_request(request, event_id)
    registration = get_object_or_404(Registration, ticket__event=event, id=registration_id)

    if registration.paid:
        messages.warning(request, "Cette place a déjà été payée.")
        return redirect(reverse("list-paid-registrations", args=[event.short_url]))
    if registration.info_state == Registration.STATE_CANNOT_PAID:
        messages.error(
            request,
            "Cette place ne peut pas être payée " "car certaines informations complémentaires n'ont pas été reçues",
        )
        return redirect(reverse("list-unpaid-registrations", args=[event.short_url]))

    if not registration.ticket.is_remain():
        if "force" in request.GET:
            LogEntry(
                user=request.user,
                event=event,
                registration=registration,
                reason=LogEntry.LOG_FORCE_LIMIT,
                operation="Paiement du billet {registration} en dépassant les limites définies",
            ).save()
        else:
            messages.error(request, "Cette place ne peut être payé car il n'en reste plus.")
            return redirect(reverse("list-unpaid-registrations", args=[event.short_url]))

    registration.paid = True
    registration.payment_date = timezone.now()
    if int(payment_method) in registration.ALLOWED_ADMIN_PAYMENT:
        registration.payment_method = int(payment_method)
    else:
        registration.payment_method = registration.PAYMENT_UNKNOW
    registration.save()

    LogEntry(
        user=request.user,
        event=event,
        registration=registration,
        reason=LogEntry.LOG_MANUAL_PAYEMENT,
        operation="Place {{registration}} marquée comme payée par {}".format(
            registration.PAYMENT_METHODS_DICT[registration.payment_method]
        ),
    ).save()

    mail_confirm_registration(registration)

    messages.success(request, "Billet payé avec succès.")
    return redirect(reverse("list-paid-registrations", args=[event.short_url]))


@manager_protected(Manager.EVENT_MANAGER)
@require_http_methods(["POST"])
def registration_oppose(request, event_id, registration_id, action=None):
    event = Event.get_event_for_request(request, event_id)
    registration = get_object_or_404(Registration, ticket__event=event, id=registration_id, paid=True)

    c = {
        "registration": registration,
        "ticket": registration.ticket,
        "event": registration.ticket.event,
    }

    if (
        registration.opposed
        and action == "remove"
        and (
            (
                (
                    registration.reason_opposed == Registration.OPPOSED_BY_ADMIN
                    or registration.reason_opposed == Registration.REFUND_BY_ADMIN
                )
                and registration.ticket.is_remain()
            )
            or "force" in request.GET
        )
    ):
        if not registration.ticket.is_remain():
            LogEntry(
                user=request.user,
                event=event,
                registration=registration,
                reason=LogEntry.LOG_FORCE_LIMIT,
                operation="Suppression de l'opposition du billet {registration} en dépassant les limites définies",
            ).save()
        raison_opposition = "inconnu"
        if registration.reason_opposed == Registration.OPPOSED_BY_ADMIN:
            raison_opposition = "admin"
        elif registration.reason_opposed == Registration.OPPOSED_BY_PAYPAL:
            raison_opposition = "paypal"
        LogEntry(
            user=request.user,
            event=event,
            registration=registration,
            reason=LogEntry.LOG_OPPOSITION_ADMIN,
            operation="Opposition sur la place {{registration}} levée (raison opposition : {})".format(
                raison_opposition
            ),
        ).save()
        registration.refund = False
        registration.opposed = False
        registration.save()
        messages.success(request, "Opposition levée avec succès.")
        if not registration.owner.is_anonyme:
            EmailMessage(
                "Billetterie INP-net : levée de l’opposition sur un de vos billets",
                render_to_string("emails/registration_unopposed.txt", c),
                settings.DEFAULT_FROM_EMAIL,
                [registration.owner.email],
                reply_to=[registration.ticket.event.email_event],
            ).send()
    elif action in ["refund", "norefund"]:
        if action == "refund":
            registration.refund = True
            email = "emails/registration_refund.txt"
            log = "Place {registration} remboursé par un admin"
            registration.reason_opposed = Registration.REFUND_BY_ADMIN
        else:
            registration.refund = False
            email = "emails/registration_opposed.txt"
            log = "Place {registration} mise en opposition par un admin"
            if registration.reason_opposed != Registration.OPPOSED_BY_PAYPAL:
                registration.reason_opposed = Registration.OPPOSED_BY_ADMIN
        LogEntry(
            user=request.user,
            event=event,
            registration=registration,
            reason=LogEntry.LOG_OPPOSITION_ADMIN,
            operation=log,
        ).save()
        registration.opposed = True
        registration.save()
        messages.success(request, "Billet mis en opposition avec succès.")
        if not registration.owner.is_anonyme:
            EmailMessage(
                "Billetterie INP-net : opposition sur un de vos billets",
                render_to_string(email, c),
                settings.DEFAULT_FROM_EMAIL,
                [registration.owner.email],
                reply_to=[registration.ticket.event.email_event],
            ).send()
    else:
        messages.error(
            request, "{}. Impossible d'enlever l'opposition".format(registration.get_reason_opposed_display())
        )

    if "details" in request.GET:
        return redirect(reverse("registration-edit", args=[event.short_url, registration.id]))
    return redirect(reverse("list-paid-registrations", args=[event.short_url]))


@manager_protected(Manager.EVENT_MANAGER)
@require_http_methods(["POST"])
def registration_email(request, event_id, registration_id):
    event = Event.get_event_for_request(request, event_id)
    registration = get_object_or_404(Registration, ticket__event=event, id=registration_id, paid=True)
    email = request.POST["email"]
    # test email validity
    try:
        validate_email(email)
    except ValidationError:
        messages.error(request, "Saisissez une adresse de courriel valide.")
    else:
        log = "Place {registration} envoyé par email à %s" % email
        mail_confirm_registration(registration, email)
        LogEntry(
            user=request.user, event=event, registration=registration, reason=LogEntry.LOG_EMAIL_BILLET, operation=log
        ).save()
        messages.success(request, "Ticket envoyé par email")
    return redirect(reverse("registration-edit", args=[event_id, registration_id]))


@login_required
def registration_list_self(request):
    registrations = Registration.objects.filter(owner=request.user)
    registrations_passed = registrations.filter(ticket__event__end_date__lt=timezone.now())
    registrations_to_come = registrations.filter(ticket__event__end_date__gte=timezone.now())

    c = {
        "registrations_passed": registrations_passed,
        "registrations_to_come": registrations_to_come,
    }

    return render(request, "guichet/registration_list.html", c)


@login_required
def access_info_sup_file(request, registration_id, info_sup_id):
    registration = get_object_or_404(Registration, id=registration_id)
    info_sup = get_object_or_404(Info_Sup_Recv_File, registration=registration, pk=info_sup_id)

    if request.user != registration.owner and not registration.ticket.event.is_manager(request.user):
        raise PermissionDenied

    if not info_sup.upload_file:
        raise Http404

    path = os.path.join(settings.MEDIA_ROOT, info_sup.upload_file.path)

    return sendfile(request, path)


@login_required
def registration_paypal_return(request, registration_id):
    # don't change ticket with this function.
    # Paypal paiement was verify with IPN
    registration = get_object_or_404(Registration, id=registration_id, owner=request.user)
    maj_state_registration(registration)  # par sécurité

    if registration.paid:
        messages.success(request, "Vous avez acheté votre place avec succès.")
        return redirect(reverse("show-registration", args=[registration.id]))
    else:
        messages.warning(
            request,
            "Votre paiement est en cours de traitement, " "veuillez actualiser la page d’ici quelques secondes.",
        )
        return redirect(reverse("show-registration", args=[registration.id]))


@login_required
def registration_paypal_cancel(request, registration_id):
    registration = get_object_or_404(Registration, id=registration_id, owner=request.user)
    maj_state_registration(registration)  # par sécurité

    return redirect(reverse("show-registration", args=[registration.id]))


@login_required
def registration_details(request, registration_id):
    registration = get_object_or_404(Registration, id=registration_id, owner=request.user)
    maj_state_registration(registration)  # par sécurité

    info_form = InfoSupReceivedForm(request.POST or None, request.FILES or None, registration=registration)
    if request.method == "POST" and info_form.is_valid():
        if not info_form.uuid_check():
            # Raced condition on info.version
            messages.error(request, "Une erreur est survenue. Merci de resaisir le formulaire.")
            return redirect(reverse("show-registration", args=[registration.id]))
        updated_info = info_form.get_updated_objects()
        for info in updated_info:
            info.save()
        if updated_info:
            messages.success(request, "Informations enregistrées.")
        return redirect(reverse("show-registration", args=[registration.id]))

    paypal_form = None
    if not registration.paid and registration.info_state != Registration.STATE_CANNOT_PAID:
        paypal_dict = {
            "business": registration.ticket.event.paypal_account.email,
            "amount": registration.get_real_amount(),
            "item_name": "Évènement « {} », place « {} »".format(
                registration.ticket.event.name, registration.ticket.name
            ),
            "invoice": "{}{}".format(settings.PAYPAL_INVOICE_PREFIX, registration.id),
            "currency_code": "EUR",
            "notify_url": settings.BASE_URL + reverse("paypal-ipn"),
            "return_url": settings.BASE_URL + reverse("paypal-return-registration", args=[registration.id]),
            "cancel_return": settings.BASE_URL + reverse("paypal-cancel-registration", args=[registration.id]),
        }
        if registration.ticket.event.paypal_account.no_account or not registration.ticket.online_payment:
            paypal_form = None
        elif registration.ticket.event.paypal_account.secure_available():
            private_cert = registration.ticket.event.paypal_account.get_cert_priv()
            public_cert = registration.ticket.event.paypal_account.get_cert_pub()
            cert_id = registration.ticket.event.paypal_account.cert_id
            paypal_form = PayPalEncryptedPaymentsForm(
                initial=paypal_dict, private_cert=private_cert, public_cert=public_cert, cert_id=cert_id
            )
        else:
            paypal_form = PayPalPaymentsForm(initial=paypal_dict)

    ticket_category = {
        Ticket.YEAR_1A: "1A",
        Ticket.YEAR_2A: "2A",
        Ticket.YEAR_3A: "3A",
        Ticket.YEAR_4A: "4A",
        Ticket.YEAR_5APLUS: "5A+",
    }
    info_state = registration.get_state_info_sup()
    info_expected_before_payement = []
    for info in info_state["missing"] + info_state["pending"] + info_state["rejected"]:
        if not info.reference.after_payement:
            info_expected_before_payement.append(info)

    # Define if a info can be give or changed
    active_info_form = False
    for info in registration.get_list_info_sup():
        if info.reference.type_info != Info_Sup.INFO_SIMPLE and (
            info.reference.info_editable or info.state != Info_Sup_Recv.STATE_ACCEPTED
        ):
            active_info_form = True
            break

    c = {
        "registration": registration,
        "dl_formats": ["pdf", "jpeg"],
        "ticket": registration.ticket,
        "event": registration.ticket.event,
        "paypal_form": paypal_form,
        "Registration": Registration,
        "info_associated_form": info_form.get_forms(),
        "active_info_form": active_info_form,
        "info_expected_before_payement": info_expected_before_payement,
    }

    if registration.lydia_request_uuid:
        c["lydia_status"] = lydia_get_status(request, registration.id)

    if registration.ticket.get_year_choice():
        c["category"] = "Billet " + ",".join([ticket_category[year] for year in registration.ticket.get_year_choice()])

    return render(request, "guichet/registration_details.html", c)


@login_required
@require_http_methods(["POST"])
def registration_cancel(request, registration_id):
    registration = get_object_or_404(Registration, id=registration_id, paid=False)
    access_right = registration.ticket.event.get_access_manager(request.user)
    if access_right < Manager.EVENT_MANAGER and registration.owner != request.user:
        raise PermissionDenied

    event = registration.ticket.event
    registration.delete()

    messages.success(request, "Réservation annulée avec succès.")

    if "manager" in request.GET:
        return redirect(reverse("list-unpaid-registrations", args=[event.short_url]))
    elif (
        event.event_confidentiality != Event.EVENT_PRIVE
        or event.at_least_one_ticket_accessible(request.user)
        or event.has_customer(request.user)
    ):  # on évite un 404 si l'utilisateur ne peux plus voir l'event
        return redirect(reverse("show-event", args=[event.short_url]))
    else:
        return redirect(reverse("list-events"))


@login_required
def registration_print(request, registration_id, format="PDF"):
    registration = get_object_or_404(Registration, id=registration_id)
    if not registration.ticket.event.is_manager(request.user) and registration.owner != request.user:
        raise PermissionDenied

    if not registration.ticket.design:
        raise Http404

    if not registration.paid or registration.info_state != Registration.STATE_CAN_ENTER:
        raise PermissionDenied

    format = format.upper()
    if format not in ["PDF", "JPEG"]:
        raise Http404

    content_types = {
        "JPEG": "image/jpeg",
        "PDF": "application/x-pdf",
    }

    img = None
    try:
        img = registration_to_image(registration, format=format)
    except BaseException:
        messages.error(
            request,
            "Une erreur s’est produite durant la création de votre billet. "
            "Veuillez réessayer plus tard. Si l’erreur persiste, "
            "veuillez contacter INP-net à l’adresse "
            '<a href="mailto:inp-net@bde.inp-toulouse.fr">inp-net@bde.inp-toulouse.fr</a>.',
        )
        return redirect(reverse("show-registration", args=[registration.id]))

    if registration.owner != request.user:
        LogEntry(
            user=request.user,
            event=registration.ticket.event,
            registration=registration,
            reason=LogEntry.LOG_PRINT_BILLET,
            operation="Impression du billet {registration} par l'interface admin",
        ).save()

    return HttpResponse(img, content_type=content_types[format])


@login_required
def registration_mail(request, registration_id):
    registration = get_object_or_404(Registration, id=registration_id)
    if not registration.ticket.event.is_manager(request.user) and registration.owner != request.user:
        raise PermissionDenied

    if not registration.paid or registration.info_state != Registration.STATE_CAN_ENTER:
        raise PermissionDenied

    try:
        mail_confirm_registration(registration)
    except BaseException:
        messages.error(
            request,
            "Une erreur s’est produite durant l’envoi de votre billet. "
            "Veuillez réessayer plus tard. Si l’erreur persiste, "
            "veuillez contacter INP-net à l’adresse "
            '<a href="mailto:inp-net@bde.inp-toulouse.fr">inp-net@bde.inp-toulouse.fr</a>.',
        )
    else:
        messages.success(request, "Votre billet vous a été envoyé par email.")

    return redirect(reverse("show-registration", args=[registration.id]))


@manager_protected(Manager.EVENT_MANAGER)
def registration_edit(request, event_id, registration_id):
    event = Event.get_event_for_request(request, event_id)
    registration = get_object_or_404(Registration, ticket__event=event, id=registration_id)

    info_sup_form = InfoSupReceivedForm(
        request.POST or None, request.FILES or None, registration=registration, admin_form=True
    )
    if request.method == "POST" and info_sup_form.is_valid():
        if not info_sup_form.uuid_check():
            # Raced condition on info.version
            messages.error(request, "Une erreur est survenue. Merci de resaisir le formulaire.")
            return redirect(reverse("registration-edit", args=[registration.ticket.event.short_url, registration.id]))
        info_recv = info_sup_form.get_updated_objects()
        for info in info_recv:
            if info.state != Info_Sup_Recv.objects.get(pk=info.pk).state:
                if info.state == Info_Sup_Recv.STATE_ACCEPTED:
                    message = "L'information {} a été acceptée pour la place {{registration}}.".format(
                        info.reference.name
                    )
                elif info.state == Info_Sup_Recv.STATE_PENDING:
                    message = "L'information {} a été mise en attente pour la place {{registration}}.".format(
                        info.reference.name
                    )
                elif info.state == Info_Sup_Recv.STATE_REJECTED:
                    message = "L'information {} a été rejetée pour la place {{registration}}.".format(
                        info.reference.name
                    )
                else:
                    message = "L'information {} a été marquée comme manquante pour la place {{registration}}.".format(
                        info.reference.name
                    )
            else:
                message = "L'information {} a été modifiée pour la place {{registration}}.".format(info.reference.name)
            LogEntry(
                user=request.user,
                event=event,
                reason=LogEntry.LOG_DOCUMENT,
                registration=registration,
                operation=message,
            ).save()
            info.save()
        registration.save()
        # gestion des billets gratuits
        maj_state_registration(registration)
        if (
            registration.get_real_amount() == 0
            and registration.info_state != Registration.STATE_CANNOT_PAID
            and not registration.paid
        ):
            registration.paid = True
            registration.payment_method = Registration.PAYMENT_UNKNOW
            registration.payment_date = timezone.now()
            registration.save()
            mail_confirm_registration(registration)
            messages.info(
                request, "Le billet a été marqué comme payé car tous les documents nécessaires ont été fournis"
            )
        messages.success(request, "Les documents ont été enregistrés")
        return redirect(reverse("registration-edit", args=[event.short_url, registration.id]))

    ipns = registration.get_ipns()
    c = {
        "info_sup_form": info_sup_form.get_forms(),
        "registration": registration,
        "dl_formats": ["pdf", "jpeg"],
        "ipns": ipns,
        "billetterie_prod": not settings.DEBUG,
        "logs": LogEntry.objects.filter(registration=registration),
        "methods": list(filter(lambda m: m[0] in Registration.ALLOWED_ADMIN_PAYMENT, Registration.PAYMENT_METHODS)),
    }

    return render(request, "guichet/admin_edit_registration.html", c)


@login_required
def manage_protected_file(request, path):
    insecure_path = os.path.join(settings.MEDIA_ROOT, "protected", path)

    if not os.path.isfile(insecure_path):
        # check if file exists but no verify that path is in media/protected
        raise Http404

    real_path = os.path.realpath(insecure_path)
    real_protected_path = os.path.realpath(os.path.join(settings.MEDIA_ROOT, "protected"))
    if not real_path.startswith(real_protected_path):
        # file not in media/protected
        raise Http404

    bdd_path = os.path.join("protected", os.path.relpath(real_path, real_protected_path))

    # check in ticket.design
    if Ticket.objects.filter(design=bdd_path).exists():
        ticket = Ticket.objects.filter(design=bdd_path).first()
        return redirect(
            reverse("show-template-ticket", args=[ticket.event.short_url, ticket.id]) + os.path.basename(bdd_path)
        )

    # check in info_sup_recv_file.upload_file
    if Info_Sup_Recv_File.objects.filter(upload_file=bdd_path).exists():
        info_sup = Info_Sup_Recv_File.objects.filter(upload_file=bdd_path).first()
        return redirect(
            reverse("access-info-sup-file", args=[info_sup.registration.id, info_sup.pk]) + os.path.basename(bdd_path)
        )

    raise Http404


@manager_required
def get_event_xlsx(request, event_id):
    event = Event.get_event_for_request(request, event_id)

    xlsx = Event_export(event).get_xlsx()
    response = HttpResponse(
        xlsx.getvalue(), content_type="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    )
    response["Content-Disposition"] = 'attachment; filename="{}-export.xlsx"'.format(event.short_url)
    return response


@manager_required
def get_event_targz(request, event_id):
    event = Event.get_event_for_request(request, event_id)

    return Event_export(event).get_tar_archive(request)


@manager_required
def get_event_zip(request, event_id):
    event = Event.get_event_for_request(request, event_id)

    return Event_export(event).get_zip_archive(request)
